@@ -4,9 +4,8 @@ import logging
 import sys
 import urllib.parse as urlparse
 import uuid
-from dataclasses import dataclass
 from pathlib import Path
-from typing import Any, Dict, List, Optional, Tuple, Union
+from typing import Any, Dict, List, Mapping, Optional, Tuple, Union
 
 import async_retriever as ar
 import cytoolz as tlz
@@ -15,7 +14,6 @@ from owslib.wfs import WebFeatureService
 from owslib.wms import WebMapService
 from pydantic import AnyHttpUrl, BaseModel, validator
 from shapely.geometry import LineString, MultiPoint, Point, Polygon
-from simplejson import JSONDecodeError
 
 from . import utils
 from .exceptions import (
@@ -34,6 +32,7 @@ handler.setFormatter(logging.Formatter(""))
 logger.handlers = [handler]
 logger.propagate = False
 DEF_CRS = "epsg:4326"
+EXPIRE = -1
 
 
 class RESTValidator(BaseModel):
@@ -50,17 +49,28 @@ class RESTValidator(BaseModel):
     outformat : str, optional
         One of the output formats offered by the selected layer. If not correct
         a list of available formats is shown, defaults to ``geojson``.
-        It defaults to ``esriSpatialRelIntersects``.
     outfields : str or list
         The output fields to be requested. Setting ``*`` as outfields requests
-        all the available fields which is the default behaviour.
+        all the available fields which is the default setting.
     crs : str, optional
         The spatial reference of the output data, defaults to EPSG:4326
     max_workers : int, optional
         Max number of simultaneous requests, default to 2. Note
         that some services might face issues when several requests are sent
         simultaneously and will return the requests partially. It's recommended
-        to avoid using too many workers unless you are certain the web service can handle it.
+        to avoid using too many workers unless you are certain the web service
+        can handle it.
+    verbose : bool, optional
+        If True, prints information about the requests and responses,
+        defaults to False.
+    disable_retry : bool, optional
+        If ``True`` in case there are any failed queries, no retrying attempts
+        is done and object IDs of the failed requests is saved to a text file
+        which its path can be accessed via ``self.failed_path``.
+    expire_after : int, optional
+        Expiration time for response caching in seconds, defaults to -1 (never expire).
+    disable_caching : bool, optional
+        If ``True``, disable caching requests, defaults to False.
     """
 
     base_url: AnyHttpUrl
@@ -68,35 +78,39 @@ class RESTValidator(BaseModel):
     outformat: str = "geojson"
     outfields: Union[List[str], str] = "*"
     crs: str = DEF_CRS
-    max_workers: int = 4
+    max_workers: int = 1
+    verbose: bool = False
+    disable_retry: bool = False
+    expire_after: float = EXPIRE
+    disable_caching: bool = False
 
     @validator("base_url")
-    def _layer_from_url(cls, v):
+    def _layer_from_url(cls, v: AnyHttpUrl) -> Tuple[str, Optional[int]]:
         return cls._split_url(urlparse.unquote(v))
 
     @validator("layer")
-    def _integer_layer(cls, v, values):
+    def _integer_layer(cls, v: Optional[int], values: Dict[str, Any]) -> int:
         if v is None:
-            lyr = values["base_url"][1]
-            if lyr is None:
+            try:
+                return int(values["base_url"][1])
+            except TypeError as ex:
                 msg = "Either layer must be passed as an argument or be included in ``base_url``"
-                raise MissingInputs(msg)
-            return lyr
+                raise MissingInputs(msg) from ex
         return v
 
     @validator("outfields")
-    def _outfields_to_list(cls, v):
+    def _outfields_to_list(cls, v: Union[List[str], str]) -> List[str]:
         return v if isinstance(v, list) else [v]
 
     @validator("crs")
-    def _valid_crs(cls, v):
+    def _valid_crs(cls, v: str) -> pyproj.CRS:
         try:
             return pyproj.CRS(v)
         except pyproj.exceptions.CRSError as ex:
             raise InvalidInputType("crs", "a valid CRS") from ex
 
     @validator("max_workers")
-    def _positive_integer_threads(cls, v):
+    def _positive_integer_threads(cls, v: int) -> int:
         if v <= 0:
             raise InvalidInputType("max_workers", "positive integer")
         return v
@@ -120,34 +134,51 @@ class ArcGISRESTfulBase:
     ----------
     base_url : str, optional
         The ArcGIS RESTful service url. The URL must either include a layer number
-        after the last ``/`` in the url or the target layer must be passed as an argument.
+        after the last ``/`` in the url or the target layer must be passed as an
+        argument.
     layer : int, optional
-        Target layer number, defaults to None. If None layer number must be included as after
-        the last ``/`` in ``base_url``.
+        Target layer number, defaults to None. If None layer number must be
+        included as after the last ``/`` in ``base_url``.
     outformat : str, optional
         One of the output formats offered by the selected layer. If not correct
         a list of available formats is shown, defaults to ``geojson``.
         It defaults to ``esriSpatialRelIntersects``.
     outfields : str or list
         The output fields to be requested. Setting ``*`` as outfields requests
-        all the available fields which is the default behaviour.
+        all the available fields which is the default setting.
     crs : str, optional
         The spatial reference of the output data, defaults to EPSG:4326
     max_workers : int, optional
         Max number of simultaneous requests, default to 2. Note
         that some services might face issues when several requests are sent
         simultaneously and will return the requests partially. It's recommended
-        to avoid using too many workers unless you are certain the web service can handle it.
+        to avoid using too many workers unless you are certain the web service
+        can handle it.
+    verbose : bool, optional
+        If True, prints information about the requests and responses,
+        defaults to False.
+    disable_retry : bool, optional
+        If ``True`` in case there are any failed queries, no retrying attempts
+        is done and object IDs of the failed requests is saved to a text file
+        which its path can be accessed via ``self.failed_path``.
+    expire_after : int, optional
+        Expiration time for response caching in seconds, defaults to -1 (never expire).
+    disable_caching : bool, optional
+        If ``True``, disable caching requests, defaults to False.
     """
 
     def __init__(
         self,
-        base_url: str,
+        base_url: AnyHttpUrl,
         layer: Optional[int] = None,
         outformat: str = "geojson",
         outfields: Union[List[str], str] = "*",
         crs: Union[str, pyproj.CRS] = DEF_CRS,
         max_workers: int = 1,
+        verbose: bool = False,
+        disable_retry: bool = False,
+        expire_after: float = EXPIRE,
+        disable_caching: bool = False,
     ) -> None:
         validated = RESTValidator(
             base_url=base_url,
@@ -156,6 +187,10 @@ class ArcGISRESTfulBase:
             outfields=outfields,
             crs=crs,
             max_workers=max_workers,
+            verbose=verbose,
+            disable_retry=disable_retry,
+            expire_after=expire_after,
+            disable_caching=disable_caching,
         )
         self.base_url = validated.base_url[0]
         self.layer = validated.layer
@@ -163,10 +198,13 @@ class ArcGISRESTfulBase:
         self.outfields = validated.outfields
         self.crs = validated.crs
         self.max_workers = validated.max_workers
+        self.verbose = validated.verbose
+        self.disable_retry = validated.disable_retry
+        self.expire_after = validated.expire_after
+        self.disable_caching = validated.disable_caching
 
         self.out_sr = pyproj.CRS(self.crs).to_epsg()
         self.n_features = 0
-        self.featureids: List[Tuple[str, ...]] = []
         self.url = f"{self.base_url}/{self.layer}"
         self.query_url = f"{self.url}/query"
         self.valid_layers: Dict[str, str] = {}
@@ -177,11 +215,12 @@ class ArcGISRESTfulBase:
         self.valid_fields: List[str] = []
         self.field_types: Dict[str, str] = {}
         self.feature_types: Optional[Dict[str, str]] = None
-        self.retry: bool = False
         self.return_m: bool = False
+        self.return_geom: bool = True
         self.n_missing: int = 0
         self.total_n_features: int = 0
-        self.failed_path: Path = Path("cache", f"{uuid.uuid4().hex}.txt")
+        self.failed_path: Optional[Path] = None
+        self.request_id: Optional[str] = None
 
         self.initialize_service()
 
@@ -200,25 +239,34 @@ class ArcGISRESTfulBase:
         if any(f not in self.valid_fields for f in self.outfields):
             raise InvalidInputValue("outfields", self.valid_fields)
 
-    def partition_oids(self, oids: Union[List[int], int, None]) -> List[Tuple[str, ...]]:
-        """Partition feature IDs based on service's max record number."""
-        if oids is None:
+    def partition_oids(self, oids: Union[List[int], int]) -> List[Tuple[str, ...]]:
+        """Partition feature IDs based on ``self.max_nrecords``."""
+        oid_list = [oids] if isinstance(oids, int) else set(oids)
+        if len(oid_list) == 0:
             raise ZeroMatched
 
-        oid_list = [str(oids)] if isinstance(oids, int) else [str(v) for v in oids]
-
         self.n_features = len(oid_list)
-        if not self.retry:
+        if not self.disable_retry and self.verbose:
             logger.info(f"Found {self.n_features:,} features in the requested region.")
-        return list(tlz.partition_all(self.max_nrecords, sorted(oid_list)))
+        return tlz.partition_all(self.max_nrecords, [str(i) for i in oid_list])  # type: ignore
 
-    def get_features(self, return_m: bool = False) -> List[Dict[str, Any]]:
+    def get_features(
+        self,
+        featureids: List[Tuple[str, ...]],
+        return_m: bool = False,
+        return_geom: bool = True,
+    ) -> List[Dict[str, Any]]:
         """Get features based on the feature IDs.
 
         Parameters
         ----------
-        return_m : bool
-            Whether to activate the Return M (measure) in the request, defaults to False.
+        featureids : list
+            List of feature IDs.
+        return_m : bool, optional
+            Whether to activate the Return M (measure) in the request,
+            defaults to ``False``.
+        return_geom : bool, optional
+            Whether to return the geometry of the feature, defaults to ``True``.
 
         Returns
         -------
@@ -228,20 +276,27 @@ class ArcGISRESTfulBase:
         payloads = [
             {
                 "objectIds": ",".join(ids),
-                "returnGeometry": "true",
+                "returnGeometry": f"{return_geom}".lower(),
                 "outSR": self.out_sr,
                 "outfields": ",".join(self.outfields),
                 "ReturnM": f"{return_m}".lower(),
                 "f": self.outformat,
             }
-            for ids in self.featureids
+            for ids in featureids
         ]
+        if self.request_id is None:
+            self.request_id = uuid.uuid4().hex
 
-        return self._get_response(payloads, "POST")
+        resp = self.get_response(self.query_url, payloads, "POST")
+
+        resp = self._cleanup_resp(resp, payloads)
+        if len(resp) == 0:
+            raise ZeroMatched
+        return resp
 
     def _set_service_properties(self) -> None:
         try:
-            rjson = ar.retrieve([self.base_url], "json", [{"params": {"f": "json"}}])[0]
+            rjson = self.get_response(self.base_url, [{"f": "json"}])[0]
             self.valid_layers = {f"{lyr['id']}": lyr["name"] for lyr in rjson["layers"]}
             self.query_formats = rjson["supportedQueryFormats"].replace(" ", "").lower().split(",")
 
@@ -253,7 +308,7 @@ class ArcGISRESTfulBase:
                     break
             self.extent = utils.match_crs(bounds, crs, DEF_CRS)
 
-        except (JSONDecodeError, KeyError) as ex:
+        except (ValueError, KeyError) as ex:
             raise ServiceError(self.base_url) from ex
 
         with contextlib.suppress(KeyError):
@@ -262,7 +317,7 @@ class ArcGISRESTfulBase:
 
     def _set_layer_properties(self) -> None:
         """Set properties of the target layer."""
-        rjson = ar.retrieve([self.url], "json", [{"params": {"f": "json"}}])[0]
+        rjson = self.get_response(self.url, [{"f": "json"}])[0]
         self.valid_fields = list(
             set(
                 utils.traverse_json(rjson, ["fields", "name"])
@@ -279,7 +334,7 @@ class ArcGISRESTfulBase:
                 zip((tlz.pluck("id", rjson["types"])), tlz.pluck("name", rjson["types"]))
             )
 
-    def _esri_query(
+    def esri_query(
         self,
         geom: Union[
             LineString,
@@ -289,7 +344,7 @@ class ArcGISRESTfulBase:
             Tuple[float, float, float, float],
         ],
         geo_crs: str = DEF_CRS,
-    ) -> Dict[str, Union[str, bytes]]:
+    ) -> Mapping[str, str]:
         """Generate geometry queries based on ESRI template."""
         geom = utils.match_crs(geom, geo_crs, self.crs)
 
@@ -310,55 +365,60 @@ class ArcGISRESTfulBase:
 
         raise InvalidInputType("geom", "LineString, Polygon, Point, MultiPoint, tuple")
 
-    def _retry(self, return_m: bool, partition_fac: float) -> List[Dict[str, Any]]:
+    def _retry(
+        self, return_m: bool, return_geo: bool, partition_fac: float
+    ) -> List[Dict[str, Any]]:
         """Retry failed requests."""
-        with open(self.failed_path) as f:
+        with open(self.failed_path) as f:  # type: ignore
             oids = [int(i) for i in f.read().splitlines()]
 
         max_nrecords = self.max_nrecords
-        self.max_nrecords = int(partition_fac * max_nrecords)
-        self.featureids = self.partition_oids(oids)
+        self.max_nrecords = int(max(partition_fac * max_nrecords, 1))
+        features = self.get_features(self.partition_oids(oids), return_m, return_geo)
         self.max_nrecords = max_nrecords
 
-        return self.get_features(return_m)
+        return features
 
-    def _retry_failed_requests(self) -> List[Dict[str, Any]]:
+    def retry_failed_requests(self) -> List[Dict[str, Any]]:
         """Retry failed requests."""
-        self.retry = True
+        retry = self.disable_retry
+        self.disable_retry = True
+        caching = self.disable_caching
+        self.disable_caching = True
+
         fac_min = 1.0 / self.max_nrecords
         n_retry = 4
         partition_fac = [(0.5 - fac_min) / (n_retry - 1) * i + fac_min for i in range(n_retry)]
         features = []
         for f in partition_fac[::-1]:
             try:
-                features.append(self._retry(self.return_m, f))
+                features.append(self._retry(self.return_m, self.return_geom, f))
             except ServiceError:
                 continue
 
-        logger.info(
-            f"Total of {self.n_missing} out of {self.total_n_features} "
-            + "features are not available on the server."
-        )
-        self.failed_path.unlink()
-        self.retry = False
+        self.disable_retry = retry
+        self.disable_caching = caching
         return list(tlz.concat(features))
 
-    def _get_response(
-        self, payloads: List[Dict[str, str]], method: str = "GET"
+    def get_response(
+        self, url: str, payloads: List[Dict[str, str]], method: str = "GET"
     ) -> List[Dict[str, Any]]:
         """Send payload and get the response."""
         req_key = "params" if method == "GET" else "data"
-        urls, kwds = zip(*((self.query_url, {req_key: p}) for p in payloads))
+        urls, kwds = zip(*((url, {req_key: p}) for p in payloads))
         try:
-            resp = ar.retrieve(
-                urls, "json", kwds, request_method=method, max_workers=self.max_workers
+            return ar.retrieve(  # type: ignore
+                urls,
+                "json",
+                list(kwds),
+                request_method=method,
+                max_workers=self.max_workers,
+                disable=self.disable_caching,
+                expire_after=self.expire_after,
             )
-        except JSONDecodeError:
+        except ValueError:
             raise ZeroMatched
 
-        resp = self._cleanup_resp(resp, payloads)
-        return resp
-
     def _cleanup_resp(
         self, resp: List[Dict[str, Any]], payloads: List[Dict[str, str]]
     ) -> List[Dict[str, Any]]:
@@ -368,38 +428,39 @@ class ArcGISRESTfulBase:
         if len(fails) > 0:
             err = resp[fails[0]]["error"]["message"]
             resp = [r for i, r in enumerate(resp) if i not in fails]
-            if len(resp) == 0 and self.retry:
+            if len(resp) == 0 and self.disable_retry:
                 raise ServiceError(err)
 
             if "objectIds" in payloads[fails[0]]:
                 oids = list(tlz.concat(payloads[i]["objectIds"].split(",") for i in fails))
                 self.n_missing = len(oids)
 
+                self.failed_path = Path("cache", f"failed_ids_{self.request_id}.txt")
                 with open(self.failed_path, "w") as f:
                     f.write("\n".join(oids))
 
-                if not self.retry:
+                if not self.disable_retry:
                     self.return_m = bool(payloads[0]["ReturnM"])
+                    self.return_geom = bool(payloads[0]["returnGeometry"])
                     self.total_n_features = self.n_features
-                    logger.info(
-                        " ".join(
-                            [
-                                f"Found {len(fails)} failed requests.",
-                                "Retrying to pluck out all available features.",
-                            ]
-                        )
-                    )
-                    resp.extend(self._retry_failed_requests())
-            else:
-                if len(resp) == 0:
-                    raise ZeroMatched
+                    if self.verbose:
+                        logger.info(f"Found {len(fails)} failed requests. Retrying ...")
+                    resp.extend(self.retry_failed_requests())
 
-                msg = (
-                    "Service failed to process some of the queries with"
-                    + "the following message, returning the rest:\n"
-                    + err
-                )
-                logger.warning(msg)
+                    if self.verbose:
+                        logger.warning(
+                            " ".join(
+                                [
+                                    f"Total of {self.n_missing} out of {self.total_n_features}",
+                                    "requested features are not available in the dataset.",
+                                    "Returning the successfully retrieved features."
+                                    "The failed object IDs have been saved in the",
+                                    f"file {self.failed_path}. The service returned the",
+                                    "following error message for the failed requests:\n",
+                                    err,
+                                ]
+                            )
+                        )
 
         return resp
 
@@ -421,8 +482,7 @@ class ArcGISRESTfulBase:
         )
 
 
-@dataclass
-class WMSBase:
+class WMSBase(BaseModel):
     """Base class for accessing a WMS service.
 
     Parameters
@@ -440,13 +500,19 @@ class WMSBase:
     crs : str, optional
         The spatial reference system to be used for requesting the data, defaults to
         epsg:4326.
+    expire_after : int, optional
+        Expiration time for response caching in seconds, defaults to -1 (never expire).
+    disable_caching : bool, optional
+        If ``True``, disable caching requests, defaults to False.
     """
 
-    url: str
+    url: AnyHttpUrl
     layers: Union[str, List[str]]
     outformat: str
     version: str = "1.3.0"
     crs: str = DEF_CRS
+    expire_after: float = EXPIRE
+    disable_caching: bool = False
 
     def __repr__(self) -> str:
         """Print the services properties."""
@@ -491,8 +557,7 @@ class WMSBase:
         return {wms[lyr].name: wms[lyr].title for lyr in list(wms.contents)}
 
 
-@dataclass
-class WFSBase:
+class WFSBase(BaseModel):
     """Base class for WFS service.
 
     Parameters
@@ -519,15 +584,21 @@ class WFSBase:
         The maximum number of records in a single request to be retrieved from the service,
         defaults to 1000. If the number of records requested is greater than this value,
         it will be split into multiple requests.
+    expire_after : int, optional
+        Expiration time for response caching in seconds, defaults to -1 (never expire).
+    disable_caching : bool, optional
+        If ``True``, disable caching requests, defaults to False.
     """
 
-    url: str
+    url: AnyHttpUrl
     layer: Optional[str] = None
     outformat: Optional[str] = None
     version: str = "2.0.0"
     crs: str = DEF_CRS
     read_method: str = "json"
     max_nrecords: int = 1000
+    expire_after: float = EXPIRE
+    disable_caching: bool = False
 
     def __repr__(self) -> str:
         """Print the services properties."""
@@ -593,17 +664,22 @@ class WFSBase:
             "typeName": self.layer,
             max_features: 1,
         }
-
-        rjson = ar.retrieve([self.url], "json", [{"params": payload}])[0]
+        rjson = ar.retrieve(
+            [self.url],
+            "json",
+            [{"params": payload}],
+            expire_after=self.expire_after,
+            disable=self.disable_caching,
+        )[0]
         valid_fields = list(
             set(
-                utils.traverse_json(rjson, ["fields", "name"])
-                + utils.traverse_json(rjson, ["fields", "alias"])
+                utils.traverse_json(rjson, ["fields", "name"])  # type: ignore
+                + utils.traverse_json(rjson, ["fields", "alias"])  # type: ignore
                 + ["*"]
             )
         )
 
         if None in valid_fields:
-            valid_fields = list(utils.traverse_json(rjson, ["features", "properties"])[0].keys())
+            valid_fields = list(utils.traverse_json(rjson, ["features", "properties"])[0].keys())  # type: ignore
 
         return valid_fields
