@@ -1,7 +1,6 @@
 """Base classes and function for REST, WMS, and WMF services."""
 import itertools
-import logging
-import sys
+import uuid
 from pathlib import Path
 from types import SimpleNamespace
 from typing import Any, Dict, List, Optional, Tuple, Union
@@ -11,25 +10,25 @@ import cytoolz as tlz
 import pyproj
 import shapely.ops as ops
 import yaml
+from pydantic import AnyHttpUrl
 from shapely.geometry import LineString, MultiPoint, MultiPolygon, Point, Polygon
 
 from . import utils
-from .core import ArcGISRESTfulBase, WFSBase, WMSBase
+from .core import DEF_CRS, EXPIRE, ArcGISRESTfulBase, WFSBase, WMSBase
 from .exceptions import InvalidInputType, InvalidInputValue, ZeroMatched
 
-logger = logging.getLogger(__name__)
-logger.setLevel(logging.INFO)
-handler = logging.StreamHandler(sys.stdout)
-handler.setFormatter(logging.Formatter(""))
-logger.handlers = [handler]
-logger.propagate = False
 
-DEF_CRS = "epsg:4326"
-
-
-class ArcGISRESTful(ArcGISRESTfulBase):
+class ArcGISRESTful:
     """Access to an ArcGIS REST service.
 
+    Notes
+    -----
+    By default, all retrieval methods retry to get the missing feature IDs,
+    if there are any. You can disable this behavior by setting ``disable_retry``
+    to ``True``. If there are any missing feature IDs after the retry,
+    they are saved to a text file, path of which can be accessed by
+    ``self.client.failed_path``.
+
     Parameters
     ----------
     base_url : str, optional
@@ -50,9 +49,47 @@ class ArcGISRESTful(ArcGISRESTfulBase):
         Number of simultaneous download, default to 1, i.e., no threading. Note
         that some services might face issues when several requests are sent
         simultaneously and will return the requests partially. It's recommended
-        to avoid performing threading unless you are certain the web service can handle it.
+        to avoid using too many workers unless you are certain the web service
+        can handle it.
+    verbose : bool, optional
+        If True, prints information about the requests and responses,
+        defaults to False.
+    disable_retry : bool, optional
+        If ``True`` in case there are any failed queries, no retrying attempts
+        is done and object IDs of the failed requests is saved to a text file
+        which its path can be accessed via ``self.client.failed_path``.
+    expire_after : int, optional
+        Expiration time for response caching in seconds, defaults to -1 (never expire).
+    disable_caching : bool, optional
+        If ``True``, disable caching requests, defaults to False.
     """
 
+    def __init__(
+        self,
+        base_url: AnyHttpUrl,
+        layer: Optional[int] = None,
+        outformat: str = "geojson",
+        outfields: Union[List[str], str] = "*",
+        crs: Union[str, pyproj.CRS] = DEF_CRS,
+        max_workers: int = 1,
+        verbose: bool = False,
+        disable_retry: bool = False,
+        expire_after: float = EXPIRE,
+        disable_caching: bool = False,
+    ) -> None:
+        self.client = ArcGISRESTfulBase(
+            base_url=base_url,
+            layer=layer,
+            outformat=outformat,
+            outfields=outfields,
+            crs=crs,
+            max_workers=max_workers,
+            verbose=verbose,
+            disable_retry=disable_retry,
+            expire_after=expire_after,
+            disable_caching=disable_caching,
+        )
+
     def oids_bygeom(
         self,
         geom: Union[
@@ -68,7 +105,7 @@ class ArcGISRESTful(ArcGISRESTfulBase):
         spatial_relation: str = "esriSpatialRelIntersects",
         sql_clause: Optional[str] = None,
         distance: Optional[int] = None,
-    ) -> None:
+    ) -> List[Tuple[str, ...]]:
         """Get feature IDs within a geometry that can be combined with a SQL where clause.
 
         Parameters
@@ -98,6 +135,11 @@ class ArcGISRESTful(ArcGISRESTfulBase):
             Valid SQL 92 WHERE clause, default to None.
         distance : int, optional
             Buffer distance in meters for the input geometries, default to None.
+
+        Returns
+        -------
+        list of tuples
+            A list of feature IDs partitioned by ``self.max_nrecords``.
         """
         valid_spatialrels = [
             "esriSpatialRelIntersects",
@@ -118,14 +160,14 @@ class ArcGISRESTful(ArcGISRESTfulBase):
         elif isinstance(geom, list) and all(len(g) == 2 for g in geom):
             geom = MultiPoint(geom)
 
-        geom_query = self._esri_query(geom, geo_crs)
+        geom_query = self.client.esri_query(geom, geo_crs)
 
         payload = {
-            **geom_query,  # type: ignore
+            **geom_query,
             "spatialRel": spatial_relation,
             "returnGeometry": "false",
             "returnIdsOnly": "true",
-            "f": self.outformat,
+            "f": self.client.outformat,
         }
         if distance:
             payload.update({"distance": f"{distance}", "units": "esriSRUnit_Meter"})
@@ -133,13 +175,15 @@ class ArcGISRESTful(ArcGISRESTfulBase):
         if sql_clause:
             payload.update({"where": sql_clause})
 
-        resp = self._get_response([payload], method="POST")[0]
+        self.client.request_id = uuid.uuid4().hex
+
+        resp = self.client.get_response(self.client.query_url, [payload], method="POST")[0]
         try:
-            self.featureids = self.partition_oids(resp["objectIds"])
+            return self.partition_oids(resp["objectIds"])
         except KeyError as ex:
             raise ZeroMatched(resp["error"]["message"]) from ex
 
-    def oids_byfield(self, field: str, ids: Union[str, List[str]]) -> None:
+    def oids_byfield(self, field: str, ids: Union[str, List[str]]) -> List[Tuple[str, ...]]:
         """Get Object IDs based on a list of field IDs.
 
         Parameters
@@ -148,19 +192,24 @@ class ArcGISRESTful(ArcGISRESTfulBase):
             Name of the target field that IDs belong to.
         ids : str or list
             A list of target ID(s).
-        """
-        if field not in self.valid_fields:
-            raise InvalidInputValue("field", self.valid_fields)
 
-        ftype = self.field_types[field]
+        Returns
+        -------
+        list of tuples
+            A list of feature IDs partitioned by ``self.max_nrecords``.
+        """
+        if field not in self.client.valid_fields:
+            raise InvalidInputValue("field", self.client.valid_fields)
+
+        ftype = self.client.field_types[field]
         if "string" in ftype:
             fids = ", ".join(f"'{i}'" for i in ids)
         else:
             fids = ", ".join(f"{i}" for i in ids)
 
-        self.oids_bysql(f"{field} IN ({fids})")
+        return self.oids_bysql(f"{field} IN ({fids})")
 
-    def oids_bysql(self, sql_clause: str) -> None:
+    def oids_bysql(self, sql_clause: str) -> List[Tuple[str, ...]]:
         """Get feature IDs using a valid SQL 92 WHERE clause.
 
         Notes
@@ -172,6 +221,11 @@ class ArcGISRESTful(ArcGISRESTfulBase):
         ----------
         sql_clause : str
             A valid SQL 92 WHERE clause.
+
+        Returns
+        -------
+        list of tuples
+            A list of feature IDs partitioned by ``self.max_nrecords``.
         """
         if not isinstance(sql_clause, str):
             raise InvalidInputType("sql_clause", "str")
@@ -180,15 +234,60 @@ class ArcGISRESTful(ArcGISRESTfulBase):
             "where": sql_clause,
             "returnGeometry": "false",
             "returnIdsOnly": "true",
-            "f": self.outformat,
+            "f": self.client.outformat,
         }
+        self.client.request_id = uuid.uuid4().hex
 
-        resp = self._get_response([payload])[0]
+        resp = self.client.get_response(self.client.query_url, [payload])[0]
         try:
-            self.featureids = self.partition_oids(resp["objectIds"])
+            return self.partition_oids(resp["objectIds"])
         except KeyError as ex:
             raise ZeroMatched(resp["error"]["message"]) from ex
 
+    def partition_oids(self, oids: Union[List[int], int]) -> List[Tuple[str, ...]]:
+        """Partition feature IDs based on ``self.max_nrecords``.
+
+        Parameters
+        ----------
+        oids : list of int or int
+            A list of feature ID(s).
+
+        Returns
+        -------
+        list of tuples
+            A list of feature IDs partitioned by ``self.max_nrecords``.
+        """
+        return self.client.partition_oids(oids)
+
+    def get_features(
+        self,
+        featureids: List[Tuple[str, ...]],
+        return_m: bool = False,
+        return_geom: bool = True,
+    ) -> List[Dict[str, Any]]:
+        """Get features based on the feature IDs.
+
+        Parameters
+        ----------
+        featureids : list
+            List of feature IDs.
+        return_m : bool, optional
+            Whether to activate the Return M (measure) in the request,
+            defaults to ``False``.
+        return_geom : bool, optional
+            Whether to return the geometry of the feature, defaults to ``True``.
+
+        Returns
+        -------
+        dict
+            (Geo)json response from the web service.
+        """
+        return self.client.get_features(featureids, return_m, return_geom)
+
+    def __repr__(self) -> str:
+        """Print the service configuration."""
+        return self.client.__repr__()
+
 
 class WMS(WMSBase):
     """Get data from a WMS service within a geometry or bounding box.
@@ -212,18 +311,32 @@ class WMS(WMSBase):
         Validate the input arguments from the WMS service, defaults to True. Set this
         to False if you are sure all the WMS settings such as layer and crs are correct
         to avoid sending extra requests.
+    expire_after : int, optional
+        Expiration time for response caching in seconds, defaults to -1 (never expire).
+    disable_caching : bool, optional
+        If ``True``, disable caching requests, defaults to False.
     """
 
     def __init__(
         self,
-        url: str,
+        url: AnyHttpUrl,
         layers: Union[str, List[str]],
         outformat: str,
         version: str = "1.3.0",
         crs: str = DEF_CRS,
         validation: bool = True,
+        expire_after: float = EXPIRE,
+        disable_caching: bool = False,
     ) -> None:
-        super().__init__(url, layers, outformat, version, crs)
+        super().__init__(
+            url=url,
+            layers=layers,
+            outformat=outformat,
+            version=version,
+            crs=crs,
+            expire_after=expire_after,
+            disable_caching=disable_caching,
+        )
 
         self.layers = [self.layers] if isinstance(self.layers, str) else self.layers
         if validation:
@@ -312,8 +425,10 @@ class WMS(WMSBase):
             "binary",
             [{"params": p} for p in payloads],
             max_workers=4,
+            expire_after=self.expire_after,
+            disable=self.disable_caching,
         )
-        return dict(zip(layers, rbinary))
+        return dict(zip(layers, rbinary))  # type: ignore
 
 
 class WFS(WFSBase):
@@ -347,11 +462,15 @@ class WFS(WFSBase):
         Validate the input arguments from the WFS service, defaults to True. Set this
         to False if you are sure all the WFS settings such as layer and crs are correct
         to avoid sending extra requests.
+    expire_after : int, optional
+        Expiration time for response caching in seconds, defaults to -1 (never expire).
+    disable_caching : bool, optional
+        If ``True``, disable caching requests, defaults to False.
     """
 
     def __init__(
         self,
-        url: str,
+        url: AnyHttpUrl,
         layer: Optional[str] = None,
         outformat: Optional[str] = None,
         version: str = "2.0.0",
@@ -359,8 +478,20 @@ class WFS(WFSBase):
         read_method: str = "json",
         max_nrecords: int = 1000,
         validation: bool = True,
+        expire_after: float = EXPIRE,
+        disable_caching: bool = False,
     ) -> None:
-        super().__init__(url, layer, outformat, version, crs, read_method, max_nrecords)
+        super().__init__(
+            url=url,
+            layer=layer,
+            outformat=outformat,
+            version=version,
+            crs=crs,
+            read_method=read_method,
+            max_nrecords=max_nrecords,
+            expire_after=expire_after,
+            disable_caching=disable_caching,
+        )
 
         if validation:
             self.validate_wfs()
@@ -410,7 +541,13 @@ class WFS(WFSBase):
             "srsName": self.crs,
         }
 
-        return ar.retrieve([self.url], self.read_method, [{"params": payload}])[0]
+        return ar.retrieve(
+            [self.url],
+            self.read_method,
+            [{"params": payload}],
+            expire_after=self.expire_after,
+            disable=self.disable_caching,
+        )[0]
 
     def getfeature_bygeom(
         self,
@@ -562,11 +699,22 @@ class WFS(WFSBase):
         }
 
         if method == "GET":
-            return ar.retrieve([self.url], self.read_method, [{"params": payload}])[0]
+            return ar.retrieve(
+                [self.url],
+                self.read_method,
+                [{"params": payload}],
+                expire_after=self.expire_after,
+                disable=self.disable_caching,
+            )[0]
 
         headers = {"content-type": "application/x-www-form-urlencoded"}
         return ar.retrieve(
-            [self.url], self.read_method, [{"data": payload, "headers": headers}], "POST"
+            [self.url],
+            self.read_method,
+            [{"data": payload, "headers": headers}],
+            "POST",
+            expire_after=self.expire_after,
+            disable=self.disable_caching,
         )[0]
 
 
