@@ -7,11 +7,10 @@ from typing import Any, Dict, List, Mapping, Optional, Tuple, TypeVar, Union
 import defusedxml.ElementTree as etree
 import pyproj
 import shapely.geometry as sgeom
-import simplejson as json
-from requests import Response
+import ujson as json
 from requests.adapters import HTTPAdapter
 from requests.exceptions import RequestException
-from requests_cache import CachedSession
+from requests_cache import CachedSession, Response
 from requests_cache.backends.sqlite import SQLiteCache
 from shapely import ops
 from urllib3 import Retry
@@ -20,7 +19,7 @@ from .exceptions import InvalidInputType, ServiceError
 
 DEF_CRS = "epsg:4326"
 BOX_ORD = "(west, south, east, north)"
-EXPIRE = 24 * 60 * 60
+EXPIRE = -1
 G = TypeVar(
     "G",
     sgeom.Point,
@@ -215,7 +214,7 @@ class ESRIGeomQuery:
     ]
     wkid: int
 
-    def point(self) -> Dict[str, Union[str, bytes]]:
+    def point(self) -> Mapping[str, str]:
         """Query for a point."""
         if not (isinstance(self.geometry, tuple) and len(self.geometry) == 2):
             raise InvalidInputType("geometry", "tuple", "(x, y)")
@@ -224,7 +223,7 @@ class ESRIGeomQuery:
         geo_json = dict(zip(("x", "y"), self.geometry))
         return self._get_payload(geo_type, geo_json)
 
-    def multipoint(self) -> Dict[str, Union[str, bytes]]:
+    def multipoint(self) -> Mapping[str, str]:
         """Query for a multi-point."""
         if not (isinstance(self.geometry, list) and all(len(g) == 2 for g in self.geometry)):
             raise InvalidInputType("geometry", "list of tuples", "[(x, y), ...]")
@@ -233,7 +232,7 @@ class ESRIGeomQuery:
         geo_json = {"points": [[x, y] for x, y in self.geometry]}
         return self._get_payload(geo_type, geo_json)
 
-    def bbox(self) -> Dict[str, Union[str, bytes]]:
+    def bbox(self) -> Mapping[str, str]:
         """Query for a bbox."""
         if not (isinstance(self.geometry, (tuple, list)) and len(self.geometry) == 4):
             raise InvalidInputType("geometry", "tuple", BOX_ORD)
@@ -242,7 +241,7 @@ class ESRIGeomQuery:
         geo_json = dict(zip(("xmin", "ymin", "xmax", "ymax"), self.geometry))
         return self._get_payload(geo_type, geo_json)
 
-    def polygon(self) -> Dict[str, Union[str, bytes]]:
+    def polygon(self) -> Mapping[str, str]:
         """Query for a polygon."""
         if not isinstance(self.geometry, sgeom.Polygon):
             raise InvalidInputType("geometry", "Polygon")
@@ -251,7 +250,7 @@ class ESRIGeomQuery:
         geo_json = {"rings": [[[x, y] for x, y in zip(*self.geometry.exterior.coords.xy)]]}
         return self._get_payload(geo_type, geo_json)
 
-    def polyline(self) -> Dict[str, Union[str, bytes]]:
+    def polyline(self) -> Mapping[str, str]:
         """Query for a polyline."""
         if not isinstance(self.geometry, sgeom.LineString):
             raise InvalidInputType("geometry", "LineString")
@@ -260,7 +259,7 @@ class ESRIGeomQuery:
         geo_json = {"paths": [[[x, y] for x, y in zip(*self.geometry.coords.xy)]]}
         return self._get_payload(geo_type, geo_json)
 
-    def _get_payload(self, geo_type: str, geo_json: Dict[str, Any]) -> Dict[str, Union[str, bytes]]:
+    def _get_payload(self, geo_type: str, geo_json: Dict[str, Any]) -> Mapping[str, str]:
         """Generate a request payload based on ESRI template.
 
         Parameters
@@ -291,7 +290,7 @@ def match_crs(geom: G, in_crs: str, out_crs: str) -> G:
     geom : list or tuple or geometry
         Input geometry which could be a list of coordinates such as ``[(x1, y1), ...]``,
         a bounding box like so ``(xmin, ymin, xmax, ymax)``, or any valid ``shapely``'s
-        geometry such as sgeom.Polygon, sgeom.MultiPolygon, etc..
+        geometry such as ``Polygon``, ``MultiPolygon``, etc..
     in_crs : str
         Spatial reference of the input geometry
     out_crs : str
@@ -299,8 +298,8 @@ def match_crs(geom: G, in_crs: str, out_crs: str) -> G:
 
     Returns
     -------
-    sgeom.LineString, sgeom.MultiLineString, sgeom.Polygon, sgeom.MultiPolygon, sgeom.Point, or sgeom.MultiPoint
-        Input geometry in the specified CRS.
+    same type as the input geometry
+        Transformed geometry in the target CRS.
 
     Examples
     --------
@@ -329,10 +328,10 @@ def match_crs(geom: G, in_crs: str, out_crs: str) -> G:
             sgeom.MultiPoint,
         ),
     ):
-        return ops.transform(project, geom)
+        return ops.transform(project, geom)  # type: ignore
 
     if isinstance(geom, tuple) and len(geom) == 4:
-        return ops.transform(project, sgeom.box(*geom)).bounds
+        return ops.transform(project, sgeom.box(*geom)).bounds  # type: ignore
 
     if isinstance(geom, list) and all(len(c) == 2 for c in geom):
         return list(zip(*project(*zip(*geom))))
@@ -490,6 +489,6 @@ def check_response(resp: str) -> str:
         return resp
     else:
         try:
-            return root[-1][0].text.strip()
+            return str(root[-1][0].text).strip()
         except IndexError:
-            return root[-1].text.strip()
+            return str(root[-1].text).strip()
